import json
import unittest
from unittest.mock import Mock, patch

from requests.auth import HTTPBasicAuth
from requests.models import Response

from gc_jira_api.Requestor import HEADERS, JIRA_BASE_ENDPOINT, RequestExecutor


class TestRequestExecutor(unittest.TestCase):
    def setUp(self):
        self.jira_username = "username"
        self.jira_password = "password"
        self.jira_server = "https://jira.example.com"
        self.executor = RequestExecutor(
            self.jira_username, self.jira_password, self.jira_server
        )

    def test_dict_to_query_string(self):
        query_dict = {"key1": "value1", "key2": "value2"}
        result = self.executor._dict_to_query_string(query_dict)
        expected = "?key1=value1&key2=value2"

        self.assertEqual(result, expected)

    @patch("gc_jira_api.Requestor.requests.get")
    def test_fetch_data(self, mock_get):
        mock_response = Mock(spec=Response)
        mock_response.json.return_value = {
            "values": [{"id": "1"}, {"id": "2"}],
            "nextPage": None,
        }
        mock_response.status_code = 200
        mock_get.return_value = mock_response

        url = "issue/createmeta"
        result = self.executor.fetch_data(url)
        expected = [{"id": "1"}, {"id": "2"}]

        self.assertEqual(result, expected)

    @patch("gc_jira_api.Requestor.requests.get")
    def test_fetch_data_absolute_url(self, mock_get):
        mock_response = Mock(spec=Response)
        mock_response.json.return_value = {
            "values": [{"id": "1"}, {"id": "2"}],
            "nextPage": None,
        }
        mock_response.status_code = 200
        mock_get.return_value = mock_response

        url = "https://jira.example.com/rest/api/3/issue/createmeta"
        result = self.executor.fetch_data(url, is_absolute_url=True)
        expected = [{"id": "1"}, {"id": "2"}]

        self.assertEqual(result, expected)

    @patch("gc_jira_api.Requestor.requests.get")
    def test_make_request_get(self, mock_get):
        mock_response = Mock(spec=Response)
        mock_response.json.return_value = {"id": "1"}
        mock_response.status_code = 200
        mock_get.return_value = mock_response

        url = f"{self.jira_server}/{JIRA_BASE_ENDPOINT}/issue/createmeta"
        result = self.executor._make_request(url, {}, 0, "GET")

        self.assertEqual(result, mock_response)

    @patch("gc_jira_api.Requestor.requests.post")
    def test_make_request_post(self, mock_post):
        mock_response = Mock(spec=Response)
        mock_response.json.return_value = {"id": "1"}
        mock_response.status_code = 201
        mock_post.return_value = mock_response

        url = f"{self.jira_server}/{JIRA_BASE_ENDPOINT}/issue"
        data = {"fields": {"summary": "Test issue"}}
        result = self.executor._make_request(url, data, 0, "POST")

        self.assertEqual(result, mock_response)
        mock_post.assert_called_once_with(
            url,
            headers=HEADERS,
            auth=HTTPBasicAuth(self.jira_username, self.jira_password),
            data=json.dumps(data),
        )

    @patch("gc_jira_api.Requestor.requests.put")
    def test_make_request_put(self, mock_put):
        mock_response = Mock(spec=Response)
        mock_response.json.return_value = {"id": "1"}
        mock_response.status_code = 200
        mock_put.return_value = mock_response

        url = f"{self.jira_server}/{JIRA_BASE_ENDPOINT}/issue/1"
        data = {"fields": {"summary": "Updated issue"}}
        result = self.executor._make_request(url, data, 0, "PUT")

        self.assertEqual(result, mock_response)
        mock_put.assert_called_once_with(
            url,
            headers=HEADERS,
            auth=HTTPBasicAuth(self.jira_username, self.jira_password),
            json=data,
        )

    @patch("gc_jira_api.Requestor.requests.delete")
    def test_make_request_delete(self, mock_delete):
        mock_response = Mock(spec=Response)
        mock_response.status_code = 204
        mock_delete.return_value = mock_response

        url = f"{self.jira_server}/{JIRA_BASE_ENDPOINT}/issue/1"
        result = self.executor._make_request(url, {}, 0, "DELETE")

        self.assertEqual(result, mock_response)
        mock_delete.assert_called_once_with(
            url,
            headers=HEADERS,
            auth=HTTPBasicAuth(self.jira_username, self.jira_password),
        )



from gc_jira_api.Requestor import RequestExecutor

POSSIBLE_DISABLE_WORDS = [
    "zz(archived)",
    "zz(archivado)",
    "zz/archivado)",
]


class JiraProject:
    def __init__(self, jira_username, jira_password, jira_server) -> None:
        self.requestor = RequestExecutor(
            jira_username,
            jira_password,
            jira_server,
        )

    def get_all_projects(self, filter_by_active=True):
        def _is_project_active(project):
            project_name_lowercase = project["name"].lower().replace(" ", "")
            for disabled_word in POSSIBLE_DISABLE_WORDS:
                if disabled_word in project_name_lowercase:
                    return False
            return True

        all_projects = self.requestor.fetch_data("project/")
        if filter_by_active:
            all_projects = list(filter(_is_project_active, all_projects))

        return all_projects

    def get_project_info(self, project_key: str):
        project_info = self.requestor.fetch_data(f"project/{project_key}/")

        project_roles = self.requestor.fetch_data(
            f"project/{project_key}/role"
        )

        project_role_members = []
        if project_roles:
            for _, role_url in project_roles.items():
                role_members = self.requestor.fetch_data(
                    role_url,
                    is_absolute_url=True,
                )
                project_role_members.append(role_members)

        return {
            "roles": project_role_members,
            "info": project_info if project_info != [] else None,
        }

    def get_project_info_by_name(self, project_name: str):
        def _get_project_data(project_info):
            project_key = project_info.get("key", None)

            project_roles = self.requestor.fetch_data(
                f"project/{project_key}/role"
            )

            project_role_members = []
            if project_roles:
                for role_name, role_url in project_roles.items():
                    if role_name == "atlassian-addons-project-access":
                        continue
                    role_members = self.requestor.fetch_data(
                        role_url,
                        is_absolute_url=True,
                    )
                    project_role_members.append(role_members)

            return {
                "content": {
                    "roles": project_role_members,
                    "info": project_info if project_info != [] else None,
                },
                "error": None,
            }

        project_info = self.requestor.fetch_data(
            f"project/search?query={project_name}&expand=lead"
        )

        if project_info and len(project_info) == 1:
            project_info = project_info[0]

            return _get_project_data(project_info=project_info)
        elif project_info and len(project_info) > 1:
            selected_project = None
            for project in project_info:
                if project["name"] == project_name:
                    selected_project = project

            if selected_project:
                return _get_project_data(project_info=project)

            return {"error": "Many projects found."}

    def get_project_members(self, project_key: str):
        project_roles = self.requestor.fetch_data(
            f"project/{project_key}/role"
        )

        project_role_members = []
        if project_roles:
            for _, role_url in project_roles.items():
                role_members = self.requestor.fetch_data(
                    role_url,
                    is_absolute_url=True,
                )
                project_role_members.append(role_members)

        return project_role_members

    def get_project_data(self, project_key: str):
        return self.requestor.fetch_data(f"project/{project_key}")

    def get_groups_for_email(self, email: str):
        user = self.requestor.fetch_data(
            "user/search",
            url_params={"query": f"{email}"},
        )

        if user:
            user_data = self.requestor.fetch_data(
                "user",
                url_params={
                    "accountId": user[0]["accountId"],
                    "expand": "groups,applicationRoles",
                },
            )

            return user_data

    def update_project(self, project_key: str, body: dict):
        return self.requestor.fetch_data(
            url=f"project/{project_key}",
            method="PUT",
            url_params=body,
        )

    def get_all_permissions_schemes(self):
        return self.requestor.fetch_data("permissionscheme")

    def get_user_info(self, username: str):
        user_info = self.requestor.fetch_data(f"user/search?query={username}")

        return user_info

    def get_user_info_by_id(self, user_id: str):
        user_info = self.requestor.fetch_data(
            f"user/search?accountId={user_id}"
        )

        return user_info

    def create_project(self, data: dict):
        created_info = self.requestor.fetch_data(
            "project",
            method="POST",
            url_params=data,
        )

        return created_info

    def get_all_project_categories(self):
        return self.requestor.fetch_data("projectCategory")

    def get_all_project_roles(self, project_code: str):
        return self.requestor.fetch_data(f"project/{project_code}/role")

    def set_users_in_project(self, project_key, role_id, request_body):
        return self.requestor.fetch_data(
            f"project/{project_key}/role/{role_id}",
            method="PUT",
            url_params=request_body,
        )

    def get_user_groups(self, user_id: str):
        return self.requestor.fetch_data(f"user/groups?accountId={user_id}")

    def bulk_issue_create(self, issues):
        def _create_issue_batches():
            batch_size = 50
            batches = []
            if issues:
                for i in range(0, len(issues), batch_size):
                    batch = issues[i : i + batch_size]  # noqa: E203
                    batches.append(batch)
            return batches

        results = []
        for issue_batch in _create_issue_batches():
            result = self.requestor.fetch_data(
                "issue/bulk",
                method="POST",
                url_params={"issueUpdates": issue_batch},
            )

            results.append(result)

        return results

    def get_custom_field_contexts(self, custom_field: str):
        return self.requestor.fetch_data(f"field/{custom_field}/contexts")

    def get_custom_field_options(self, custom_field: str, context_id: int):
        return self.requestor.fetch_data(
            f"field/{custom_field}/context/{context_id}/option"
        )

    def create_custom_field_options(
        self, custom_field: str, context_id: int, data: dict
    ):
        return self.requestor.fetch_data(
            f"field/{custom_field}/context/{context_id}/option",
            method="POST",
            url_params=data,
        )

    def get_issue_by_id(self, issue_id: str):
        return self.requestor.fetch_data(f"issue/{issue_id}")

    def get_all_project_issues(self, project_key: str):
        all_issues = []
        start_at = 0
        max_results = 100

        while True:
            response = self.requestor.fetch_data(
                f"search?jql=project={project_key}&maxResults={max_results}&startAt={start_at}"  # noqa: E501
            )

            issues = response.get("issues", [])
            all_issues.extend(issues)

            if start_at + max_results >= response.get("total", 0):
                break  # Ya se recuperaron todos los resultados

            start_at += max_results

        return {"issues": all_issues}

    def get_issue_worklogs(
        self, issue_id: str, starter_after=None, starter_before=None
    ):
        return self.requestor.fetch_data(
            f"issue/{issue_id}/worklog?startedAfter={starter_after}&startedBefore={starter_before}",  # noqa: E501
        )



import json
import logging
import time

import requests
from requests.auth import HTTPBasicAuth

JIRA_BASE_ENDPOINT = "rest/api/3"
MAX_RETRIES = 3
BACKOFF_FACTOR = 2
HEADERS = {
    "Accept": "application/json",
    "Content-Type": "application/json",
}


class RequestExecutor:
    def __init__(self, jira_username, jira_password, jira_server):
        self.jira_server = jira_server

        self.auth = HTTPBasicAuth(jira_username, jira_password)

    def _dict_to_query_string(self, dictionary: dict):
        query_string = "?"

        for key, value in dictionary.items():
            query_string += f"{key}={value}&"

        if query_string[-1] == "&":
            query_string = query_string[:-1]

        return query_string

    def fetch_data(
        self,
        url: str,
        url_params: dict = {},
        is_absolute_url=False,
        method="GET",
    ):
        if not is_absolute_url:
            url = f"{self.jira_server}/{JIRA_BASE_ENDPOINT}/{url}"

        next_url = url
        all_records = []
        while next_url:
            response = self._make_request(next_url, url_params, 0, method)
            if response:
                data = response.json()
                if isinstance(data, dict) and "values" in data.keys():
                    all_records.extend(data["values"])

                    if "nextPage" in data:
                        next_url = data["nextPage"]

                    elif "isLast" in data and not data["isLast"]:
                        start_at = data.get("startAt", 0) + data.get(
                            "maxResults", 0
                        )
                        next_url = f"{url}?startAt={start_at}"

                    else:
                        next_url = None

                else:
                    next_url = None
                    all_records = data
            else:
                next_url = None

        return all_records

    def _make_request(self, url, url_params, retry_count, method):
        try:
            if method == "PUT":
                response = requests.put(
                    url,
                    headers=HEADERS,
                    # timeout=10,
                    auth=self.auth,
                    json=url_params,
                )
            elif method == "POST":
                response = requests.post(
                    url,
                    headers=HEADERS,
                    # timeout=10,
                    auth=self.auth,
                    # data=url_params,
                    data=json.dumps(url_params),
                )
            elif method == "DELETE":
                response = requests.delete(
                    url,
                    headers=HEADERS,
                    # timeout=10,
                    auth=self.auth,
                    # data=url_params,
                    # data=json.dumps(url_params),
                )
            else:
                response = requests.get(
                    url,
                    headers=HEADERS,
                    # timeout=10,
                    auth=self.auth,
                    params=url_params,
                )

            if response.status_code == 400:
                logging.error(response.json())

            response.raise_for_status()

            return response
        except (
            requests.exceptions.Timeout,
            requests.exceptions.TooManyRedirects,
            requests.exceptions.RequestException
        ) as e:
            if retry_count >= MAX_RETRIES:
                logging.error(
                    f"[ERROR - _make_request]: Request failed after {MAX_RETRIES} attempts due to: {e}"  # noqa: E501
                )

                return None
            next_seconds = BACKOFF_FACTOR**retry_count
            logging.warning(
                f"[ERROR - _make_request]: Request timeout or too many redirects, retrying in {next_seconds} seconds ({url})"  # noqa: E501
            )
            time.sleep(next_seconds)

            return self._make_request(url, url_params, retry_count + 1, method)



import unittest
from unittest.mock import patch

from gc_jira_api.jira import JiraProject
from gc_jira_api.Requestor import RequestExecutor

JIRA_SERVER = "JIRA_SERVER_TEST"
JIRA_USERNAME = "JIRA_USERNAME_TEST"
JIRA_PASSWORD = "JIRA_PASSWORD_TEST"


class TestJiraProject(unittest.TestCase):
    @patch.object(RequestExecutor, "fetch_data")
    def test_get_all_projects(self, mock_fetch_data):
        mock_fetch_data.return_value = [
            {"name": "Project 1"},
            {"name": "zz(archived) Project 2"},
            {"name": "Active Project"},
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        all_projects = jira_project.get_all_projects(filter_by_active=True)
        self.assertEqual(len(all_projects), 2)
        self.assertEqual(all_projects[0]["name"], "Project 1")
        self.assertEqual(all_projects[1]["name"], "Active Project")

        all_projects = jira_project.get_all_projects(filter_by_active=False)
        self.assertEqual(len(all_projects), 3)

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_project_info(self, mock_fetch_data):
        mock_fetch_data.side_effect = [
            {"name": "Project Info"},  # First call return
            {"role1": "role_url1", "role2": "role_url2"},  # Second call return
            {"members": ["User1", "User2"]},  # Third call return
            {"members": ["User3", "User4"]},  # Fourth call return
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        project_info = jira_project.get_project_info("PROJECT_KEY")
        self.assertIn("info", project_info)
        self.assertIn("roles", project_info)
        self.assertEqual(project_info["info"]["name"], "Project Info")
        self.assertEqual(len(project_info["roles"]), 2)

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_project_info_by_name(self, mock_fetch_data):
        mock_fetch_data.side_effect = [
            [
                {"key": "PROJECT_KEY", "name": "Project Name"}
            ],  # First call return
            {"role1": "role_url1", "role2": "role_url2"},  # Second call return
            {"members": ["User1", "User2"]},  # Third call return
            {"members": ["User3", "User4"]},  # Fourth call return
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        project_info = jira_project.get_project_info_by_name("Project Name")
        self.assertIn("content", project_info)
        self.assertIn("roles", project_info["content"])
        self.assertEqual(len(project_info["content"]["roles"]), 2)

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_project_members(self, mock_fetch_data):
        mock_fetch_data.side_effect = [
            {"role1": "role_url1", "role2": "role_url2"},  # First call return
            {"members": ["User1", "User2"]},  # Second call return
            {"members": ["User3", "User4"]},  # Third call return
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        project_members = jira_project.get_project_members("PROJECT_KEY")
        self.assertEqual(len(project_members), 2)
        self.assertEqual(project_members[0]["members"], ["User1", "User2"])
        self.assertEqual(project_members[1]["members"], ["User3", "User4"])

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_project_data(self, mock_fetch_data):
        mock_fetch_data.return_value = {
            "key": "PROJECT_KEY",
            "name": "Project Name",
        }

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        project_data = jira_project.get_project_data("PROJECT_KEY")
        self.assertEqual(project_data["name"], "Project Name")

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_groups_for_email(self, mock_fetch_data):
        mock_fetch_data.side_effect = [
            [{"accountId": "user123"}],  # First call return
            {"groups": ["group1", "group2"], "applicationRoles": ["role1"]},
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        user_data = jira_project.get_groups_for_email("user@example.com")
        self.assertIn("groups", user_data)
        self.assertIn("applicationRoles", user_data)
        self.assertEqual(user_data["groups"], ["group1", "group2"])

    @patch.object(RequestExecutor, "fetch_data")
    def test_update_project(self, mock_fetch_data):
        mock_fetch_data.return_value = {
            "key": "PROJECT_KEY",
            "name": "Updated Project",
        }

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        updated_project = jira_project.update_project(
            "PROJECT_KEY", {"name": "Updated Project"}
        )
        self.assertEqual(updated_project["name"], "Updated Project")

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_all_permissions_schemes(self, mock_fetch_data):
        mock_fetch_data.return_value = [
            {"id": "1", "name": "Scheme1"},
            {"id": "2", "name": "Scheme2"},
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        permissions_schemes = jira_project.get_all_permissions_schemes()
        self.assertEqual(len(permissions_schemes), 2)
        self.assertEqual(permissions_schemes[0]["name"], "Scheme1")

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_user_info(self, mock_fetch_data):
        mock_fetch_data.return_value = [
            {"accountId": "user123", "displayName": "User One"}
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        user_info = jira_project.get_user_info("User One")
        self.assertEqual(user_info[0]["displayName"], "User One")

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_user_info_by_id(self, mock_fetch_data):
        mock_fetch_data.return_value = [
            {"accountId": "user123", "displayName": "User One"}
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        user_info = jira_project.get_user_info_by_id("user123")
        self.assertEqual(user_info[0]["displayName"], "User One")

    @patch.object(RequestExecutor, "fetch_data")
    def test_create_project(self, mock_fetch_data):
        mock_fetch_data.return_value = {
            "key": "NEW_PROJECT",
            "name": "New Project",
        }

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        created_project = jira_project.create_project({"name": "New Project"})
        self.assertEqual(created_project["name"], "New Project")

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_all_project_categories(self, mock_fetch_data):
        mock_fetch_data.return_value = [
            {"id": "1", "name": "Category1"},
            {"id": "2", "name": "Category2"},
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        categories = jira_project.get_all_project_categories()
        self.assertEqual(len(categories), 2)
        self.assertEqual(categories[0]["name"], "Category1")

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_all_project_roles(self, mock_fetch_data):
        mock_fetch_data.return_value = {
            "role1": "role_url1",
            "role2": "role_url2",
        }

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        roles = jira_project.get_all_project_roles("PROJECT_CODE")
        self.assertEqual(len(roles), 2)
        self.assertIn("role1", roles)

    @patch.object(RequestExecutor, "fetch_data")
    def test_set_users_in_project(self, mock_fetch_data):
        mock_fetch_data.return_value = {
            "roleId": "123",
            "projectKey": "PROJECT_KEY",
        }

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        response = jira_project.set_users_in_project(
            "PROJECT_KEY", "123", {"users": ["user1", "user2"]}
        )
        self.assertEqual(response["roleId"], "123")

    @patch.object(RequestExecutor, "fetch_data")
    def test_get_user_groups(self, mock_fetch_data):
        mock_fetch_data.return_value = [
            {"groupName": "Group1"},
            {"groupName": "Group2"},
        ]

        jira_project = JiraProject(JIRA_USERNAME, JIRA_PASSWORD, JIRA_SERVER)

        user_groups = jira_project.get_user_groups("user123")
        self.assertEqual(len(user_groups), 2)
        self.assertEqual(user_groups[0]["groupName"], "Group1")
